package nl.uvt.locator;

import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.util.Enumeration;
import java.util.Properties;

import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;

import nu.xom.Builder;
import nu.xom.ParsingException;
import nu.xom.ValidityException;
import nu.xom.Document;

public class HttpConnection {
	
	public static Document httpGetDocument( String url, Properties args )
	throws UnsupportedEncodingException, MalformedURLException, IOException, ValidityException, ParsingException {
		Document responseDocument = null;
		InputStream is = HttpGetInputStream(url, args);
		Builder builder = new Builder();
		responseDocument =  builder.build(is);
		return responseDocument;
	}

	public static Document httpGetDocument( String url )
	throws UnsupportedEncodingException, MalformedURLException, IOException, ValidityException, ParsingException {
		return httpGetDocument( url, null ); 
	}

	/**
	 * Perform a HTTP GET request and return InputStream.
	 * Based on: Jason Hunter & William Crawford, Java Servlet Programming, 2nd Edition,
	 * pp 316-319, O'Reilly
	 * 
	 * @param url URL of web service
	 * @param args key/value pairs
	 * @return InputStream with response
	 */

	public static String createUrlString( String base, Properties attVals ) throws UnsupportedEncodingException{
		String url = base;
		String argString = "";
		if ( attVals != null) {
			argString = "?" + toEncodedString(attVals);
			url += argString;
		}
		return url;
	}

	public static InputStream HttpGetInputStream( String url, Properties props ) 
	throws UnsupportedEncodingException,
	MalformedURLException,
	IOException {
		URL fullUrl = new URL( url );
		URLConnection con = fullUrl.openConnection();
		if ( props != null ){
			String useCache = props.getProperty("useCache", "false");
			if ( useCache.equalsIgnoreCase("true") )
				con.setUseCaches(true);
			else
				con.setUseCaches(false);
			String keepAlive = props.getProperty("Keep-Alive", "300");
			con.setRequestProperty("Keep-Alive", keepAlive  );
			String timeout = props.getProperty("urlTimeout", "10000"); // take max 10 seconds, or what is configured
			con.setReadTimeout( Integer.parseInt(timeout) );
		}
		return con.getInputStream();
	}

	/**
	 * Transforms key/value pairs into query string
	 * 
	 * @param args Series of key/value pairs
	 * @return String like 'a=1&b=2', without question mark 
	 */

	private static String toEncodedString(Properties args )
	throws UnsupportedEncodingException {
		return toEncodedString( args, "UTF-8" );
	}
	
	private static String toEncodedString(Properties args, String enc )
	throws UnsupportedEncodingException {
		StringBuffer buf = new StringBuffer();
		Enumeration names = args.propertyNames();
		while (names.hasMoreElements()) {
			String name = (String) names.nextElement();
			String value = args.getProperty(name);
			buf.append(URLEncoder.encode(name, enc ) + "=" + URLEncoder.encode(value, enc ));
			if (names.hasMoreElements())
				buf.append("&");
		}

		return buf.toString();
	}

	public static Document httpGetHTMLDocument( String url, Properties httpProps )
	throws UnsupportedEncodingException, MalformedURLException, IOException, ValidityException, ParsingException, SAXException {
		nu.xom.Document responseDocument = null;
		InputStream is = HttpGetInputStream(url, httpProps);
		XMLReader tagsoup = XMLReaderFactory.createXMLReader("org.ccil.cowan.tagsoup.Parser");
		Builder builder = new Builder( tagsoup );
		responseDocument =  builder.build(is);
		is.close();
		return responseDocument;
	}

}
