package nl.uvt.locator;

import java.io.*;
import java.net.*;
import java.math.*;
import java.security.*;
import java.util.*;
import java.util.regex.*;

import nu.xom.*;

public class LinkDbInfo extends LinkInfo {

	List<Threshold> thresholds;

	LinkDbInfo(){
		thresholds = new LinkedList<Threshold>();
	}

	public boolean inRange( QueryData pQ) throws LocatorException{
		return Threshold.inRange( thresholds, pQ );
	}


	void fill( Node rec ){
		String res = getValue( rec, "parameters/issn");
		if ( res.length() == 0 ){
			res = getValue( rec, "issn");		
		}
		if ( res.length() > 0 ){
			Pattern issnPat = Pattern.compile("([\\dX]*)");
			Matcher issnMatcher = issnPat.matcher(res);
			if ( issnMatcher.find() ){
				linkIssn = issnMatcher.group();
//				logger.debug( "matched!"+issn);
			}
		}
		res = getValue( rec, "title" );
		if ( res.length() > 0 )
			linkTitle = res;
		res = getValue( rec, "url");
		if ( res.length() > 0 )
			linkUrl = res;
		res = getValue( rec, "template" );
		if ( res.length() > 0 )
			linkMethod = res;
		res = getValue( rec, "publisher");
		if ( res.length() > 0 )
			linkPublisher = res;
		Nodes nodes = rec.query( "threshold ");
		for ( int i=0; i < nodes.size(); ++i ){
			Threshold th = new Threshold();
			th.fill( nodes.get(i) );
			thresholds.add( th );
		}
	}

	public String toString(){
		String result = "LinkDbInfo:";
		result += " issn='" + linkIssn;
		result += "' title='" + linkTitle;
		result += "' url='" + linkUrl;
		result += "' template='" + linkMethod;
		result += "' publisher='" + linkPublisher;
		result += "' thresholds=";
		for ( int i=0; i < thresholds.size(); ++i ){
			result += thresholds.get(i).toString();
		}
		return result;
	}

	public boolean applyMethod( QueryData pQ ){
		if ( isDOImethod( linkMethod ) ){
			logger.debug("Method" + linkMethod + " is a well-known DOI method");
			String myUrl = generateDOI( pQ );
			if ( myUrl.equals("") ){
				rank = 3;
				myUrl = linkUrl;
			}
			else
				rank = 1;
			generatedUrl = myUrl;			
		}
		else  {
			logger.debug("Method= " + linkMethod );
			if ( linkMethod.equals("ABI") ) {
				String newUri = "http://gateway.proquest.com/openurl?ctx_ver=Z39.88-2003&res_id=xri:pqd&rft_val_fmt_=ori:fmt:kev:mtx:journal&issn=";
				newUri += linkIssn;
				newUri += "&jtitle=" + linkTitle;
				if ( pQ.atitle.length() > 0 ){
					try {
						newUri += "&atitle=" + URLEncoder.encode( pQ.atitle, "ISO-8859-1" );
					} catch (UnsupportedEncodingException e) {
						logger.fatal( "encoding title failed", e );
					}
				}
				if ( linkVolume > 0 )
					newUri += "&volume=" + linkVolume;
				else
					if ( pQ.volume > 0 )
						newUri += "&volume=" + pQ.volume;
				if ( !pQ.oIssue.equals( "" ) ){
					newUri += "&issue=" + pQ.oIssue;
				}
				else if ( linkIssue > 0 )
					newUri += "&issue=" + linkIssue;
				else
					if ( pQ.issue > 0 )
						newUri += "&issue=" + pQ.issue;
				if ( !pQ.spage.equals("") )
					newUri += "&spage=" + pQ.spage;
				rank = 1;
				generatedUrl = newUri;
			}
			else if ( linkMethod.equals("AMA") ) {
				String newUri = linkUrl + "/cgi/reprint/";
				newUri += getRequired( "volume", linkVolume, pQ.volume );
				newUri += "/" + getRequired( "issue", linkIssue, pQ.issue );
				if ( !pQ.spage.equals("") )
					newUri += "/" + pQ.spage;
				rank = 1;
				generatedUrl = newUri;
			}
			else if ( linkMethod.equals("APP") ) {
				String newUri = linkUrl + "/cgi/content/full";
				if ( pQ.volume >= 0 )
					newUri += "/" + pQ.volume;
				if ( pQ.issue >= 0 )
					newUri += "/" + pQ.issue;
				if ( !pQ.spage.equals("") )
					newUri += "/" + pQ.spage;
				rank = 1;
				generatedUrl = newUri;
			}
			else if ( linkMethod.equals("Berkeley") ) {
				generatedUrl = linkUrl;
				generatedUrl += "/vol" + getRequired( "volume", linkVolume, pQ.volume );
				generatedUrl += "/iss" + getRequired( "issue", linkIssue, pQ.issue );
				rank = 1;
			}
			else if ( linkMethod.equals("Bernouilli") ) {
				rank = 3;
				generatedUrl = linkUrl;
			}
			else if ( linkMethod.equals("catchword") ) {
				String newUri = "http://www.ingentaselect.com/rpsv/cgi-bin/cgi?body=linker&reqidx=";
				newUri += getRequired( "issn", linkIssn, pQ.issn );
				newUri += "(" + getRequired( "year", pQ.year) + ")L.";
				if ( !pQ.spage.equals("") )
					newUri += pQ.spage;
				rank = 1;
				generatedUrl = newUri;
			}
			else if ( linkMethod.equals("Clark") ) {
				generatedUrl = linkUrl;
				generatedUrl += "issue.cfm?issue=" + pQ.year + "_0";
				generatedUrl +=  getRequired("issue", linkIssue, pQ.issue );
				generatedUrl += ".html"; 
				rank = 1;
			}
			else if ( linkMethod.equals("DOAJ") ) {
				rank = 3;
				generatedUrl = linkUrl;
			}
			else if ( linkMethod.equals("Euclid") ) {
				rank = 3;
				generatedUrl = linkUrl;
			}
			else if ( linkMethod.equals("Hart") ) {
				rank = 1;
				generatedUrl = linkUrl;
				generatedUrl += "/volumes/" + getRequired("volume", linkVolume, pQ.volume );
				generatedUrl += "/issues/" + getRequired("issue", linkIssue, pQ.issue );
			}
			else if ( linkMethod.equals("HeinOnline") ) {
				rank = 3;
				generatedUrl = linkUrl;
			}
			else if ( linkMethod.equals("Highwire") ) {
				rank = 1;
				generatedUrl = linkUrl + "/cgi/content/full/";
				generatedUrl += getRequired("volume", linkVolume, pQ.volume );
				generatedUrl += "/" + getRequired("issue", linkIssue, pQ.issue );
				generatedUrl += "/" + getRequired("spage", pQ.spage );
			}
			else if ( linkMethod.equals("IBDF") ) {
				rank = 3;
				generatedUrl = linkUrl;
			}
			else if ( linkMethod.equals("IngentaConnect") ) {
				String openURL = "http://openurl.ingenta.com/content?";
				String myIssn = getRequired( "issn", linkIssn, pQ.issn );
				if ( !myIssn.equals("") ){
					myIssn = myIssn.substring(0, 4) + "-" + myIssn.substring(4,8);
				}
				if ( !pQ.genre.equals("") )
					openURL += "?genre=" + pQ.genre;
				openURL += "&issn=" + myIssn;
				if ( linkIssue > 0  )
					openURL += "&issue=" + linkIssue;
				else if ( pQ.issue > 0 )
					openURL += "&issue=" + pQ.issue;
				if ( linkVolume > 0  )
					openURL += "&volume=" + linkVolume;
				else if ( pQ.volume > 0 )
					openURL += "&volume=" + pQ.volume;
				if ( !pQ.spage.equals("") )
					openURL += "&spage=" + pQ.spage;
				rank = 1;
				generatedUrl = openURL;
			}
			else if ( linkMethod.equals("IPO") ) {
				rank = 3;
				generatedUrl = linkUrl;
			}
			else if ( linkMethod.equals("Journal")  ) {
				rank = 3;
				generatedUrl = linkUrl;
			}
			else if ( linkMethod.equals("JSTOR")  ) {
				String myUrl = "http://links.jstor.org/sici?";
				myUrl += "sici=" + getRequired("issn", linkIssn );
				myUrl += "(" + getRequired("year", pQ.year ) + ")";
				if ( linkVolume > 0 )
					myUrl += linkVolume;
				else
					if ( pQ.volume > 0 )
						myUrl += pQ.volume;
				if ( linkIssue > 0 )
					myUrl += ":" + linkIssue;
				else
					if ( pQ.issue > 0 )
						myUrl += ":" + pQ.issue;
				myUrl += "%3C" + getRequired( "spage", pQ.spage ) + "%3E2.0.CO;2-B";
				rank = 1;
				generatedUrl = myUrl;
			}
			else if ( linkMethod.equals("KLI")  ) {
				generatedUrl = "http://www.kluwerlawonline.com/toc.php?area=Journals&mode=bypub&level=6&values=Journals~~";
				generatedUrl += getRequired( "title", linkTitle, pQ.title );
				generatedUrl += "~Volume+" + getRequired( "volume", pQ.volume );
				generatedUrl += "+(" + getRequired( "year", pQ.year );
				generatedUrl += ")~Issue+" + getRequired( "issue", pQ.issue );
				rank = 1;
			}
			else if ( linkMethod.equals("Login-Hint")  ) {
				rank = 3;
				generatedUrl = linkUrl;
			}
			else if ( linkMethod.equals("Metapress")  ) {
				String openURL = "http://www.metapress.com/openurl.asp";
				if ( !pQ.genre.equals("") )
					openURL += "?genre=" + pQ.genre;
				if ( !linkIssn.equals("") )
					openURL += "&issn=" + linkIssn;
				else if ( !pQ.issn.equals("") )
					openURL += "&issn=" + pQ.issn;
				if ( linkIssue > 0  )
					openURL += "&issue=" + linkIssue;
				else if ( pQ.issue > 0 )
					openURL += "&issue=" + pQ.issue;
				if ( linkVolume > 0  )
					openURL += "&volume=" + linkVolume;
				else if ( pQ.volume > 0 )
					openURL += "&volume=" + pQ.volume;
				if ( !pQ.spage.equals("") )
					openURL += "&spage=" + pQ.spage;
				rank = 1;
				generatedUrl = openURL;
			}
			else if ( linkMethod.equals("MLCV")  ) {
				String spage = lPad( pQ.spage, '0', 4 );
				String volume = lPad( pQ.volume, '0', 3 );
				String part = linkUrl;
				if ( part.length() > 36 ){
					part = part.substring(36);
					String[] splits = part.split("/");
					if ( splits.length > 1 ){
						rank = 1;
						generatedUrl = "http://www.multilingual-matters.net/" + splits[0] + "/" 
						+ volume + "/" + spage + "/" + splits[0] + volume + spage + ".pdf";
					}
				}
			}	
			else if ( linkMethod.equals("NASW") ) {
				rank = 3;
				generatedUrl = linkUrl;
			}
			else if ( linkMethod.equals("NBLC") ) {
				rank = 1;
				generatedUrl = "http://www.leesgoedplus.nl/archief/";
				generatedUrl += getRequired("title", linkTitle ) + "/";
				generatedUrl += getRequired("year", pQ.year ) + "/";
				generatedUrl += getRequired("issue", linkIssue, pQ.issue ) + ".pdf";
			}
			else if ( linkMethod.equals("NCTE") ) {
				rank = 3;
				generatedUrl = linkUrl;
			}
			else if ( linkMethod.equals("OLC") ) {
				// OLC is not handled by us!
				// should link to the docserver
				diagnostics.add( new Diagnostic( 1, "method OLC is not handled by the Locator", "docserver link needed"));
			}
			else if ( linkMethod.equals("ovid") ) {
				String newUri = "http://gateway.ovid.com/ovidweb.cgi?T=JS&PAGE=fulltext&D=ovft&MODE=ovid&NEWS=N&SEARCH=";
				newUri += linkIssn.substring(0, 4) + "-" + linkIssn.substring(4,8);
				newUri += ".is+and+";
				newUri += getRequired( "volume", linkVolume, pQ.volume );
				newUri += ".vo+and+";
				newUri += getRequired( "issue", linkIssue, pQ.issue );
				newUri += ".ip+and+" + getRequired( "spage", pQ.spage );
				if ( diagnostics.isEmpty() ){
					newUri += ".pg&LOGOUT=n";
					rank = 1;
					generatedUrl = newUri;
				}
			}
			else if ( linkMethod.equals("Peeters") ) {
				int pos = linkUrl.indexOf("code=");
				String code = linkUrl.substring(pos+5);
				String newUri = "http://poj.peeters-leuven.be/issue.php?journal_code=" + code;
				newUri += "&issue=" + getRequired( "issue", linkIssue, pQ.issue);
				newUri += "&vol=" + getRequired( "volume", linkVolume, pQ.volume);
				rank = 1;
				generatedUrl = newUri;
			}
			else if ( linkMethod.equals("PPP") ) {
				rank = 3;
				generatedUrl = linkUrl;
			}
			else if ( linkMethod.equals("RCP") ) {
				rank = 3;
				generatedUrl = linkUrl;
			}
			else if ( linkMethod.equals("SAGE") ) {
				// String generatedUrl = generateOpenURL( pQ );
				generatedUrl = linkUrl + "/cgi/reprint/"	
				+ getRequired( "volume", pQ.volume ) + "/" 
				+ getRequired( "issue", pQ.issue ) + "/" 
				+ getRequired( "spage", pQ.spage ) + ".pdf";
				rank = 1;
			}
			else if ( linkMethod.equals("SD")){
				String myUrl = generateDOI( pQ );
				if ( myUrl.equals("") ){
					logger.debug( "no DOI, try harder:");
					if ( !(linkVolume == 0 && pQ.volume == 0 )
							&& !pQ.spage.equals("") ){
						logger.debug( "create md5 hashed SD link");
						String SDorigin="UTILBURG";
						String SDsalt = "(CxZ:=Jrxs;hXYWS!KVpdj!jgeO7NSdJ";
						String volKey = linkIssn;
						volKey += "#" + getRequired("volume", linkVolume, pQ.volume );
						volKey += "#" + pQ.spage;
						if ( pQ.issue >= 0 )
							volKey += "#" + pQ.issue;
						String md5source = "_ob=GatewayURL&_origin=" + SDorigin 
						+ "&_method=citationSearch&_volkey=" + volKey
						+ "&_version=1" + SDsalt;
//						logger.debug("md5:'" + md5source + "'" );
						MessageDigest md5;
						String md5Hash = "";
						try {
							md5 = MessageDigest.getInstance("MD5");
							md5.update(md5source.getBytes(), 0, md5source.length());
							byte mdBytes[] = md5.digest();
							md5Hash = new BigInteger(1,mdBytes).toString(16);
							if ( md5Hash.length() < 32 )
								md5Hash = '0' + md5Hash; 
//							logger.debug("md5 hash:'" + md5Hash + "'" );
						} catch (NoSuchAlgorithmException e) {
							logger.fatal("MD5 hasher", e);
						}
						try {
							myUrl = "http://www.sciencedirect.com/science?_ob=GatewayURL&_origin="
								+ SDorigin + "&_method=citationSearch&_volkey="
								+ URLEncoder.encode(volKey, "ISO-8859-1" ) + "&_version=1&md5=" + md5Hash;
						} catch (UnsupportedEncodingException e) {
							logger.fatal( "encoding failed", e );
						}
						rank = 2;
					}
					else {
						logger.debug( "last resort: linkDb link");
						rank = 3;
						myUrl = linkUrl;
					}
				}
				else {
					rank = 1;
				}
				generatedUrl = myUrl;
			}
			else if ( linkMethod.equals("springerlink") ) {
				// springerlink does openUrl, but the linkDb link already gives genre and issn
				String openURL = linkUrl;
				openURL += "&volume=" + getRequired( "volume", pQ.volume );
				openURL += "&issue=" + getRequired( "issue", pQ.issue );
				openURL += "&spage=" + getRequired( "spage", pQ.spage );
				generatedUrl = openURL;
				rank = 1;
			}
			else if ( linkMethod.equals("swetsnet") ) {
				String newUri = "http://www.swetsnet.nl/link/access_db?issn=";
				newUri += linkIssn;
				newUri += "&vol=";
				if ( linkVolume > 0 )
					newUri += linkVolume;
				else
					newUri += pQ.volume; 
				newUri += "&iss=";
				if ( linkIssue > 0 )
					newUri += linkIssue;
				else
					newUri += pQ.issue;
				newUri += "&page=" +pQ.spage;
				newUri += "&year=" + pQ.year + "&FT=1";
				rank = 1;
				generatedUrl = newUri;
			}
			else if ( linkMethod.equals("TF")  ) {
				rank = 1;
				generatedUrl = "http://www.informaworld.com/openurl?" + generateOpenURL( pQ );
			}
			else if ( linkMethod.equals("VenR")  ) {
				rank = 3;
				generatedUrl = linkUrl;
			}
			else if ( linkMethod.equals("WESTLAW") ) {
				rank = 3;
				generatedUrl = linkUrl;
			}
			else if ( linkMethod.equals("WL") ) {
				rank = 3;
				generatedUrl = linkUrl;
				generatedUrl += "/PasswordPDF/Vol" + getRequired("volume", linkVolume, pQ.volume );
				generatedUrl += "no" + getRequired( "issue", linkIssue, pQ.issue );
				generatedUrl += "_" + getRequired( "year", pQ.year );
			}
			else if ( linkMethod.equals("WPNR") ) {
				rank = 3;
				generatedUrl = linkUrl;

			}
			else {
				// default handling is DOI
				logger.debug( "try DOI" );
				String myUrl = generateDOI( pQ );
				if ( myUrl.equals("") ){
					rank = 3;
					myUrl = linkUrl;
				}
				else {
					logger.info("the method " + linkMethod + " gave a DOI, maybe it's a good advise to add it to DOImethods in your config?");
					rank = 1;
				}
				generatedUrl = myUrl;
			}
		}
		if ( !diagnostics.isEmpty() ){
			logger.warn("Diagnostics: " );
			for( int i=0; i < diagnostics.size(); ++i )
				logger.warn( diagnostics.get(i).toString() );
			generatedUrl = "";
		}
		logger.debug("generated URL: "+ generatedUrl );
		return generatedUrl != "";
	}

	static Set<String> DOImethods = null;

	private boolean isDOImethod(String linkMethod) {
		if ( DOImethods == null ){
			DOImethods = new HashSet<String>();
			String doiM = Locator.locatorConfig.getProperty("DOImethods","");
			String [] doiS = doiM.split(",");
			for ( int i=0; i < doiS.length; ++i ){
				DOImethods.add(doiS[i].trim());
			}
			logger.debug("Known DOI methods: " + DOImethods.toString() );
		}
		if ( DOImethods.contains(linkMethod) )
			return true;
		else
			return false;
	}

	private String lPad(String string, char c, int len ) {
		// pad string with c to get length len
		String result = string;
		while ( result.length() < len ) result = c + result;
		return result;
	}

	private String lPad( Integer val, char c, int len ) {
		String valS = Integer.toString(val);
		return lPad( valS, c, len );
	}

	private int getRequired(String string, int val ) {
		if	( val >= 0 )
			return val;
		else
			diagnostics.add(misPar(string) );
		return -1;
	}


	private int getRequired(String string, int lVal, int qVal ) {
		if	( lVal >=  0 )
			return lVal;
		else if ( qVal >= 0 )
			return qVal;
		else 
			diagnostics.add(misPar(string) );
		return -1;	
	}

	private String getRequired(String string, String val ) {
		if	( !val.equals("") )
			return val;
		else
			diagnostics.add(misPar(string) );
		return "";
	}

	private String getRequired(String string, String lVal, String qVal ) {
		if	( !lVal.equals("") )
			return lVal;
		else if ( !qVal.equals("") )
			return qVal;
		else
			diagnostics.add(misPar(string) );
		return "";
	}


	private Diagnostic misPar( String par ){
		return new Diagnostic( 10, "missing required parameter", par );
	}


	private String generateDOI( QueryData pQ ){
		String crossrefUri = Locator.locatorConfig.getProperty("crossrefUri");
		try {
			Properties prop = new Properties();
			String qData = "";
			qData += linkIssn + "|";
//			if ( linkTitle.length() > 0 )
//			qData +=  URLEncoder.encode(linkTitle, "ISO-8859-1" ) +"||";
//			else
			qData += pQ.atitle +"||";
			if ( linkVolume > 0 )
				qData += linkVolume + "|";
			else
				qData += pQ.volume + "|";
			if ( linkIssue > 0 )
				qData += linkIssue + "|";
			else if ( pQ.issue > 0 )
				qData += pQ.issue + "|";
			else
				qData += "|";
			qData += pQ.spage + "|";
			qData += pQ.year + "|||";
			prop.setProperty("qdata", qData );
			prop.setProperty( "pwd", "crpw021");
			prop.setProperty( "usr", "tuli");
			String url = HttpConnection.createUrlString(crossrefUri, prop); 
			InputStream result = HttpConnection.HttpGetInputStream( url, null );
			BufferedReader in = new BufferedReader( new InputStreamReader(result) );
			logger.debug( "crossref returns:");
			String doi = "";
			while ( in.ready() ){
				String line = in.readLine();
				if ( line.contains("|") ){
					String [] arr = line.split( "\\|" );
					if ( arr.length > 8 && arr[9] != "" )
						doi = "http://dx.doi.org/" +  URLEncoder.encode(arr[9], "ISO-8859-1" );
				}
				logger.debug( line );
			}
			return doi;
		}
		catch (Exception e) 
		{
			logger.debug( "crossref failed, uri" );
			return "";
		}
	}


}
