package nl.uvt.locator;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import nu.xom.*;

public class MerescoInfo extends LinkInfo {
	String epn;
	String status;
	String callNumber;
	String material;
	String collection;
	String action;
	List<ThRange> ranges; 

	private class ThRange {
		List<Threshold> rangeItems;
		ThRange(){
			rangeItems = new LinkedList<Threshold>();
		}

		public String toString(){
			String s = "[";
			for( int i=0; i<rangeItems.size(); ++i ){
				s += rangeItems.get(i).toString();
				if ( i < rangeItems.size()-1 )
					s += ",";
			}
			s += "]";
			return s;
		}
	}

	MerescoInfo(){
		epn = "";
		status = "";
		callNumber =  "";
		material = "";
		action = "";
		collection = "";
		ranges = new LinkedList<ThRange>();
	}

	static XPathContext modsContext = Locator.sruContext;

	public static final String modsNamespace = "http://www.loc.gov/mods/v3";
	public static final String schemaNamespace = "http://www.w3.org/2001/XMLSchema-instance";
	public static final String extNamespace = "http://drcwww.uvt.nl/~place/search%20engine/extension";

	public static void fillMultipleRecord(Node inputRecord, List<LinkInfo> meResult) {
		Locator.logger.debug( "split one record per occurrence");
//		Locator.logger.debug( "input record: " + inputRecord.toXML() );
		MerescoInfo data = new MerescoInfo();
		data.linkMethod = "meresco";
		Nodes nodes= inputRecord.query("./mods:recordInfo/mods:recordIdentifier", MerescoInfo.modsContext );
		if ( nodes.size() > 0 ){
			if ( nodes.get(0).getChildCount() > 0 ) 
				data.linkIssn = nodes.get(0).getChild(0).getValue().toString();
		}
		nodes = inputRecord.query("./mods:titleInfo/mods:title", MerescoInfo.modsContext);
		if ( nodes.size() > 0 ){
			if ( nodes.get(0).getChildCount() > 0 ) 
				data.linkTitle = nodes.get(0).getChild(0).getValue().toString();
		}
		nodes = inputRecord.query("./mods:originInfo/mods:publisher", MerescoInfo.modsContext );
		if ( nodes.size() > 0 ){
			if ( nodes.get(0).getChildCount() > 0 ) 
				data.linkPublisher = nodes.get(0).getChild(0).getValue().toString();
		}
		nodes = inputRecord.query("./mods:location", MerescoInfo.modsContext )	;
		if ( nodes.size() > 0 ){
			if ( nodes.get(0).getChildCount() > 0 ) {
				//
				// occurrence can be found as a note in holdingSimple or as a tag in holdingExternal
				// there is no guarantee that occurrences are present in both, are continuous or whatever
				// therefore, we just collect all found 'occurrences' in a String set and try them all
				//
				Nodes occNodes = nodes.get(0).query("./mods:holdingSimple/mods:copyInformation/mods:note[@type='occurrence']", MerescoInfo.modsContext );
				// search the simpelHolding
				SortedSet<String> occs = new TreeSet<String>();
				if ( occNodes.size() > 0){
					Locator.logger.debug("there are " + occNodes.size() + " occurence nodes in Simple" );
					for ( int i=0; i < occNodes.size(); ++i ){
//						Locator.logger.debug("adding occ = " + occNodes.get(i).getValue() );
						occs.add(occNodes.get(i).getValue());
					}
				}
				occNodes = nodes.get(0).query("./mods:holdingExternal/a:copies/a:copy/a:occurrence", MerescoInfo.modsContext );
				// search the external Holding
				if ( occNodes.size() > 0){
					Locator.logger.debug("there are " + occNodes.size() + " occurence nodes in External" );
					for ( int i=0; i < occNodes.size(); ++i ){
//						Locator.logger.debug("adding occ = " + occNodes.get(i).getValue() );
						occs.add(occNodes.get(i).getValue());
					}
				}

				// now loop all occurences. First try External. then Simple.
				Iterator<String> occIt = occs.iterator();
				while ( occIt.hasNext() ){
					String occS = occIt.next();
					Locator.logger.debug("handling occ="+ occS );
					Nodes extHolding = nodes.get(0).query("./mods:holdingExternal/a:copies/a:copy[a:occurrence='"+ occS +"']", MerescoInfo.modsContext);
					if ( extHolding.size() == 0 ){
						Locator.logger.debug( "found no externalHolding for occ=" + occS );
						Nodes simpleHolding = nodes.get(0).query("./mods:holdingSimple/mods:copyInformation[mods:note[@type='occurrence' and .='"+occS+"']]", MerescoInfo.modsContext);
						if ( simpleHolding.size() == 0 )
							Locator.logger.debug("Also found NO simpleHolding!");
						else {
							Node holding = simpleHolding.get(0);
							Locator.logger.debug( "but found simpleHolding for occ=" + occS );
//							Locator.logger.debug( holding.toXML());							
							MerescoInfo rec = new MerescoInfo();
							rec.action = "";
							rec.linkIssn = data.linkIssn;
							rec.linkTitle = data.linkTitle;
							rec.linkPublisher = data.linkPublisher;
							Nodes notes = holding.query("./mods:note[@type='epn']", MerescoInfo.modsContext);
							if( notes.size() > 0 ){
								String epn = notes.get(0).getValue().toString();
								rec.epn = epn;
								Locator.logger.debug("got epn:" + epn );
							}
							notes = holding.query("./mods:note[@type='status']", MerescoInfo.modsContext);
							if( notes.size() > 0 ){
								String status = notes.get(0).getValue().toString();
								rec.status = status;
								Locator.logger.debug("got status: " + status );
							}
							notes = holding.query("./mods:shelfLocator", MerescoInfo.modsContext );
							if( notes.size() > 0 ){
								String call = notes.get(0).getValue().toString();
								Locator.logger.debug("got call: " + call );
								fixCall( call, rec );
							}
							meResult.add( rec );
						}
					}
					else {
						Node holding = extHolding.get(0);
						Locator.logger.debug( "found externalHolding for occ=" + occS );
						Locator.logger.debug( holding.toXML());
						MerescoInfo rec = new MerescoInfo();
						rec.action = "";
						rec.linkIssn = data.linkIssn;
						rec.linkTitle = data.linkTitle;
						rec.linkPublisher = data.linkPublisher;
						Nodes info = holding.query("./a:epn", MerescoInfo.modsContext );
						if ( info.size() > 0 ){
							if ( info.get(0).getChildCount() > 0 ) 
								rec.epn = info.get(0).getChild(0).getValue().toString();
						}
						info = holding.query("./a:collection", MerescoInfo.modsContext);
						if ( info.size() > 0 ){
							if ( info.get(0).getChildCount() > 0 ) 
								rec.collection = info.get(0).getChild(0).getValue().toString();
						}
						info = holding.query("./mods:enumerationAndChronology", MerescoInfo.modsContext);
						if ( info.size() > 0 ){
							String annotation = "";
							if ( info.get(0).getChildCount() > 0 ) 
								annotation = info.get(0).getChild(0).getValue().toString();
							Locator.logger.debug("got call: " + annotation );
							fixCall( annotation, rec );

						}
						info = holding.query("./a:status", MerescoInfo.modsContext);
						if ( info.size() > 0 ){
							if ( info.get(0).getChildCount() > 0 ) 
								rec.status = info.get(0).getChild(0).getValue().toString();
						}
						info = holding.query("./mods:shelfLocator", MerescoInfo.modsContext);
						if ( info.size() > 0 ){
							if ( info.get(0).getChildCount() > 0 ) 
								rec.callNumber = info.get(0).getChild(0).getValue().toString();
						}
						info = holding.query( "./a:range", MerescoInfo.modsContext);
						for ( int j=0; j < info.size(); ++j ){
							Locator.logger.debug("get range:"+j);
							ThRange th = data.new ThRange();
							Nodes thNodes = info.get(j).query( "./a:threshold", MerescoInfo.modsContext);
							for ( int k=0; k < thNodes.size(); ++k ){
								Threshold thSub = new Threshold();
								thSub.fill( thNodes.get(k), "a", MerescoInfo.modsContext );
								th.rangeItems.add( thSub );
							}
//							Locator.logger.debug("read:" + th.toString());
							rec.ranges.add( th );
						}
						meResult.add( rec );
					}
				}
			}
		}
		else {
			Locator.logger.debug( "Did not find usefull Holding Info in input record: " + inputRecord.toXML() );	
		}
	}


	private static void fixCall( String call, MerescoInfo rec){
		Locator.logger.debug("fixup " + call );
		if ( !call.equals("") )
			rec.collection = tryToFixCollection( call );
		if ( rec.collection.equals("") )
			rec.callNumber = call;
		else {
			// for now, we leave the collection in the call, to stay backward compatible
//			call = call.replace( rec.collection + " ", "" );
			rec.callNumber = call;
		}
		Locator.logger.debug( "rec.collection="+rec.collection );
		Locator.logger.debug( "rec.callNumber="+ rec.callNumber );
	}	

	static Set<String> knownCollections = null;

	static private boolean isKnownCollection( String collection ) {
		if ( knownCollections == null ){
			knownCollections = new HashSet<String>();
			String colls = Locator.locatorConfig.getProperty("knownCollections","");
			if ( colls.equals("") ){
				// hardwired default
				Locator.logger.warn("PLEASE FIX THIS: NO knownCollections found in configfile." );
				colls = "CBM, CIF, BRA, DISC, ECO, GES, JUR, LET, SOC, TFL";
				Locator.logger.warn("PLEASE FIX THIS: we use hardcoded default: " + colls );
			}
			String [] doiS = colls.split(",");
			for ( int i=0; i < doiS.length; ++i ){
				knownCollections.add(doiS[i].trim());
			}
			Locator.logger.debug("Known collections: " + knownCollections.toString() );
		}
		return knownCollections.contains( collection );
	}

	private static String tryToFixCollection(String call ) {
		//
		// try to guess the collection from the callnumber
		// rather naive
		//
		String[] parts = call.split(" ");
		for ( int i=0; i < parts.length; ++i ){
			String coll = parts[i].replaceAll("[()]","");
			if ( isKnownCollection( coll ) )
				return coll;
			if ( isKnownCollection( parts[i] ) )
				return parts[i];
		}
		return "";	
	}

	void fill( Node rec ){
		logger.debug( "fill from record:" + rec.toXML() );
		linkMethod = "meresco";
		Nodes nodes= rec.query("./mods:recordInfo/mods:recordIdentifier", modsContext);
		if ( nodes.size() > 0 ){
			if ( nodes.get(0).getChildCount() > 0 ) 
				linkIssn = nodes.get(0).getChild(0).getValue().toString();
		}
		nodes = rec.query("./mods:titleInfo/mods:title", modsContext);
		if ( nodes.size() > 0 ){
			if ( nodes.get(0).getChildCount() > 0 ) 
				linkTitle = nodes.get(0).getChild(0).getValue().toString();
		}
		nodes = rec.query("./mods:location/mods:url", modsContext);
		if ( nodes.size() > 0 ){
			if ( nodes.get(0).getChildCount() > 0 ) 
				linkUrl = nodes.get(0).getChild(0).getValue().toString();
		}
		nodes = rec.query("./mods:originInfo/mods:publisher", modsContext);
		if ( nodes.size() > 0 ){
			if ( nodes.get(0).getChildCount() > 0 ) 
				linkPublisher = nodes.get(0).getChild(0).getValue().toString();
		}
	}		

	Node toRecordNode(){
		Element record = new Element( "record" );
		record.appendChild( makeNode( "linkIssn", linkIssn ) );
		record.appendChild( makeNode( "linkTitle", linkTitle ) );
		record.appendChild( makeNode( "linkUrl", linkUrl ) );
		record.appendChild( makeNode( "generatedUrl", generatedUrl ) );
		record.appendChild( makeNode( "linkMethod", linkMethod ) );
		record.appendChild( makeNode( "action", action ) );
		record.appendChild( makeNode( "collection", collection ) );
		record.appendChild( makeNode( "epn", epn ) );
		record.appendChild( makeNode( "callNumber", callNumber ) );
		record.appendChild( makeNode( "material", material ) );
		record.appendChild( makeNode( "status", status ) );
		return record;
	}

	public boolean inRange( QueryData pQ) {
		try {
			if ( ranges.size() == 0 )
				return true;
			for ( int i=0; i < ranges.size(); ++i ){
				if ( Threshold.inRange( ranges.get(i).rangeItems, pQ ) )
					return true;
			}
			return false;
		} catch (LocatorException e) {
			logger.debug("too bad");
			return false;
		}
	}

	public boolean applyMethod( QueryData pQ ){
		logger.debug("handle Meresco method: " + status );
		logger.debug("action="+action);
		logger.debug("status="+status);
		if ( !inRange( pQ ) ){
			logger.debug( "rejected because of threshold");
			return false;
		}
		else {
			logger.debug( "collection=" + collection );
			logger.debug( "epn=" + epn );
			logger.debug( "status=" + status );
			logger.debug( "callNumber=" + callNumber );
			if ( action.equals("") || action.equals("passed")){
				if ( !linkUrl.equals("") ){
					action = "url";
					generatedUrl = linkUrl;				
				}
				else if ( !epn.equals("") && !status.equals("") && 
						!callNumber.equals("") && !collection.equals("")) {
					action = "ok";
				}
				else
					action = "reject";
			}
			if ( !action.equals("reject") ){
				logger.debug("accepted");
				return true;
			}
			else {
				logger.debug("rejected");
				return false;
			}
		}
	}

}
